@@ -30,7 +30,6 @@
 #include <qt_windows.h>
 #endif
 #ifdef Q_OS_LINUX
-#include <sys/poll.h>
 #include <sys/prctl.h>
 #endif
 
@@ -46,9 +45,6 @@
 #include <QFile>
 #include <QGuiApplication>
 #include <QLibraryInfo>
-#include <QOffscreenSurface>
-#include <QOpenGLContext>
-#include <QOpenGLFunctions>
 #include <QStandardPaths>
 #include <QThread>
 
@@ -66,10 +62,7 @@
 #undef Q_GLOBAL_STATIC
 
 using namespace std::chrono_literals;
-
-#ifdef Q_OS_LINUX
-static QByteArray s_socketpath;
-#endif
+using namespace Qt::StringLiterals;
 
 struct Args {
     Args() = default;
@@ -125,6 +118,7 @@ static KCrash::CrashFlags s_flags = KCrash::CrashFlags();
 static int s_launchDrKonqi = -1; // -1=initial value 0=disabled 1=enabled
 static int s_originalSignal = -1;
 static QByteArray s_metadataPath;
+static const bool s_crashInHandler = qEnvironmentVariableIntValue("KCRASH_CRASH_IN_HANDLER") == 1;
 
 static std::unique_ptr<char[]> s_kcrashErrorMessage;
 
@@ -132,28 +126,12 @@ namespace
 {
 const KCrash::CoreConfig s_coreConfig;
 
-std::unique_ptr<char[]> s_glRenderer; // the GL_RENDERER
 std::unique_ptr<char[]> s_qtVersion;
 
-QString glRenderer()
-{
-    QOpenGLContext context;
-    QOffscreenSurface surface;
-    surface.create();
-
-    if (!context.create()) {
-        return {};
-    }
-
-    if (!context.makeCurrent(&surface)) {
-        return {};
-    }
-
-    auto done = qScopeGuard([&context] {
-        context.doneCurrent();
-    });
-    return QString::fromUtf8(reinterpret_cast<const char *>(context.functions()->glGetString(GL_RENDERER)));
-}
+using DetailsHash = QHash<QByteArray, QByteArray>;
+std::unique_ptr<const DetailsHash> s_tags; // Sentry tags
+std::unique_ptr<const DetailsHash> s_extraData; // Sentry extra data
+std::unique_ptr<const DetailsHash> s_gpuContext; // Sentry gpu context
 
 QString bootId()
 {
@@ -196,6 +174,9 @@ LONG WINAPI win32UnhandledExceptionFilter(_EXCEPTION_POINTERS *exceptionInfo);
 static bool shouldWriteMetadataToDisk()
 {
 #ifdef Q_OS_LINUX
+    if (QStandardPaths::isTestModeEnabled()) {
+        return true;
+    }
     // NB: The daemon being currently running must not be a condition here. If something crashes during logout
     // the daemon may already be gone but we'll still want to deal with the crash on next login!
     // Similar reasoning applies to not checking the presence of the launcher socket.
@@ -235,29 +216,25 @@ void KCrash::initialize()
         const QString path = QCoreApplication::applicationFilePath();
         s_appFilePath.reset(qstrdup(qPrintable(path))); // This intentionally cannot be changed by the application!
         KCrash::setApplicationFilePath(path);
-        if (qobject_cast<QGuiApplication *>(QCoreApplication::instance())) {
-            s_glRenderer.reset(qstrdup(glRenderer().toUtf8().constData()));
-        }
     } else {
         qWarning() << "This process needs a QCoreApplication instance in order to use KCrash";
     }
 
-#ifdef Q_OS_LINUX
-    // Create socket path to transfer ptrace scope and open connection
-    s_socketpath = QFile::encodeName(QStringLiteral("%1/kcrash_%2").arg(QStandardPaths::writableLocation(QStandardPaths::RuntimeLocation)).arg(getpid()));
-#endif
-
     if (shouldWriteMetadataToDisk()) {
         // We do not actively clean up metadata via KCrash but some other service. This potentially means we litter
         // a lot -> put the metadata in a subdir.
         // This data is consumed by DrKonqi in combination with coredumpd metadata.
         const QString metadataDir = QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QStringLiteral("/kcrash-metadata");
         if (QDir().mkpath(metadataDir)) {
-            const auto bootId = ::bootId();
-            const auto exe = QString::fromUtf8(s_appName.get());
-            const auto pid = QString::number(QCoreApplication::applicationPid());
-            s_metadataPath = QFile::encodeName(metadataDir + //
-                                               QStringLiteral("/%1.%2.%3.ini").arg(exe, bootId, pid));
+            if (QStandardPaths::isTestModeEnabled()) {
+                s_metadataPath = QFile::encodeName(metadataDir + QStringLiteral("/test.ini"));
+            } else {
+                const auto bootId = ::bootId();
+                const auto exe = QString::fromUtf8(s_appName.get());
+                const auto pid = QString::number(QCoreApplication::applicationPid());
+                s_metadataPath = QFile::encodeName(metadataDir + //
+                                                   QStringLiteral("/%1.%2.%3.ini").arg(exe, bootId, pid));
+            }
         }
         if (!s_crashHandler) {
             // Always enable the default handler. We cannot create the metadata ahead of time since we do not know
@@ -498,8 +475,42 @@ void KCrash::defaultCrashHandler(int sig)
         if (!s_appFilePath) {
             fprintf(stderr, "KCrash: appFilePath points to nullptr!\n");
         } else if (ini.isWritable()) {
+            if (s_tags) {
+                // [KCrashTags]
+                ini.startTagsGroup();
+                // Add our dynamic details. Note that since we only add them to the ini they do not count against our static argv limit in the Metadata class.
+                for (const auto &[key, value] : s_tags->asKeyValueRange()) {
+                    ini.add(key.constData(), value.constData(), MetadataWriter::BoolValue::No);
+                }
+            }
+
+            if (s_extraData) {
+                // [KCrashExtra]
+                ini.startExtraGroup();
+                // Add our dynamic details. Note that since we only add them to the ini they do not count against our static argv limit in the Metadata class.
+                for (const auto &[key, value] : s_extraData->asKeyValueRange()) {
+                    ini.add(key.constData(), value.constData(), MetadataWriter::BoolValue::No);
+                }
+            }
+            if (s_kcrashErrorMessage) {
+                // And also our legacy error message
+                ini.add("--_kcrash_ErrorMessage", s_kcrashErrorMessage.get(), MetadataWriter::BoolValue::No);
+            }
+
+            if (s_gpuContext) {
+                // [KCrashGPU]
+                ini.startGPUGroup();
+                // Add our dynamic details. Note that since we only add them to the ini they do not count against our static argv limit in the Metadata class.
+                for (const auto &[key, value] : s_gpuContext->asKeyValueRange()) {
+                    ini.add(key.constData(), value.constData(), MetadataWriter::BoolValue::No);
+                }
+            }
+
+            // [KCrash]
+            ini.startKCrashGroup();
             // Add the canonical exe path so the coredump daemon has more data points to map metadata to journald entry.
             ini.add("--exe", s_appFilePath.get(), MetadataWriter::BoolValue::No);
+
             data.setAdditionalWriter(&ini);
         }
 #endif
@@ -513,10 +524,6 @@ void KCrash::defaultCrashHandler(int sig)
             }
         }
 
-        if (s_glRenderer) {
-            data.add("--glrenderer", s_glRenderer.get());
-        }
-
         if (s_qtVersion) {
             data.add("--qtversion", s_qtVersion.get());
         }
@@ -536,9 +543,12 @@ void KCrash::defaultCrashHandler(int sig)
         if (platformName == QByteArrayLiteral("xcb")) {
             // start up on the correct display
             char *display = nullptr;
-            if (auto disp = qGuiApp->nativeInterface<QNativeInterface::QX11Application>()->display()) {
-                display = XDisplayString(disp);
-            } else {
+            if (qGuiApp) {
+                if (auto disp = qGuiApp->nativeInterface<QNativeInterface::QX11Application>()->display()) {
+                    display = XDisplayString(disp);
+                }
+            }
+            if (!display) {
                 display = getenv("DISPLAY");
             }
             data.add("--display", display);
@@ -594,6 +604,13 @@ void KCrash::defaultCrashHandler(int sig)
         data.add("--thread", threadId);
 #endif
 
+        if (s_crashInHandler) {
+#if !defined(Q_OS_WIN)
+            kill(getpid(), SIGSEGV);
+#endif
+            return; // we may continue running a couple ms after the segv. return so we don't close the data!
+        }
+
         data.close();
         const int argc = data.argc;
         const char **argv = data.argv.data();
@@ -707,14 +724,6 @@ LONG WINAPI KCrash::win32UnhandledExceptionFilter(_EXCEPTION_POINTERS *exception
 
 static pid_t startDirectly(const char *argv[]);
 
-#ifdef Q_OS_LINUX
-static int write_socket(int sock, char *buffer, int len);
-static int read_socket(int sock, char *buffer, int len);
-
-static int openDrKonqiSocket(const QByteArray &socketpath);
-static int pollDrKonqiSocket(pid_t pid, int sockfd);
-#endif
-
 void KCrash::startProcess(int argc, const char *argv[], bool waitAndExit)
 {
     Q_UNUSED(argc);
@@ -732,23 +741,10 @@ void KCrash::startProcess(int argc, const char *argv[], bool waitAndExit)
         // and only send SIGCONT when it is about to attach a debugger.
 #ifdef Q_OS_LINUX
         // Declare the process that will be debugging the crashed KDE app (#245529).
-        // For now that will be DrKonqi, which may ask to transfer the ptrace scope to
-        // a debugger it is not an ancestor of (because it was started via kdeinit or
-        // KProcess::startDetached()) using a socket.
 #ifndef PR_SET_PTRACER
 #define PR_SET_PTRACER 0x59616d61
 #endif
         prctl(PR_SET_PTRACER, pid, 0, 0, 0);
-
-        int sockfd = openDrKonqiSocket(s_socketpath);
-
-        if (sockfd >= 0) {
-            // Wait while DrKonqi is running and the socket connection exists
-            // If the process was started directly, use waitpid(), as it's a child...
-            while ((running = waitpid(pid, nullptr, WNOHANG) != pid) && pollDrKonqiSocket(pid, sockfd) >= 0) { }
-            close(sockfd);
-            unlink(s_socketpath.constData());
-        }
 #endif
         if (running) {
             // If the process was started directly, use waitpid(), as it's a child...
@@ -798,143 +794,42 @@ static pid_t startDirectly(const char *argv[])
         return pid;
     }
 }
-
-#ifdef Q_OS_LINUX
-
-/*
- * Write 'len' bytes from 'buffer' into 'sock'.
- * returns 0 on success, -1 on failure.
- */
-static int write_socket(int sock, char *buffer, int len)
-{
-    ssize_t result;
-    int bytes_left = len;
-    while (bytes_left > 0) {
-        result = write(sock, buffer, bytes_left);
-        if (result > 0) {
-            buffer += result;
-            bytes_left -= result;
-        } else if (result == 0) {
-            return -1;
-        } else if ((result == -1) && (errno != EINTR) && (errno != EAGAIN)) {
-            return -1;
-        }
-    }
-    return 0;
-}
-
-/*
- * Read 'len' bytes from 'sock' into 'buffer'.
- * returns 0 on success, -1 on failure.
- */
-static int read_socket(int sock, char *buffer, int len)
-{
-    ssize_t result;
-    int bytes_left = len;
-    while (bytes_left > 0) {
-        result = read(sock, buffer, bytes_left);
-        if (result > 0) {
-            buffer += result;
-            bytes_left -= result;
-        } else if (result == 0) {
-            return -1;
-        } else if ((result == -1) && (errno != EINTR) && (errno != EAGAIN)) {
-            return -1;
-        }
-    }
-    return 0;
-}
-
-static int openDrKonqiSocket(const QByteArray &socketpath)
-{
-    int sockfd = socket(PF_UNIX, SOCK_STREAM, 0);
-    if (sockfd < 0) {
-        perror("Warning: socket() for communication with DrKonqi failed");
-        return -1;
-    }
-
-    struct sockaddr_un drkonqi_server;
-    drkonqi_server.sun_family = AF_UNIX;
-
-    if (socketpath.size() >= static_cast<int>(sizeof(drkonqi_server.sun_path))) {
-        fprintf(stderr, "Warning: socket path is too long\n");
-        close(sockfd);
-        return -1;
-    }
-    strcpy(drkonqi_server.sun_path, socketpath.constData());
-
-    unlink(drkonqi_server.sun_path); // remove potential stale socket
-    if (bind(sockfd, (struct sockaddr *)&drkonqi_server, sizeof(drkonqi_server)) < 0) {
-        perror("Warning: bind() for communication with DrKonqi failed");
-        close(sockfd);
-        unlink(drkonqi_server.sun_path);
-        return -1;
-    }
-
-    listen(sockfd, 1);
-
-    return sockfd;
-}
-
-static int pollDrKonqiSocket(pid_t pid, int sockfd)
-{
-    struct pollfd fd;
-    fd.fd = sockfd;
-    fd.events = POLLIN;
-    int r;
-    do {
-        r = poll(&fd, 1, 1000); // wait for 1 second for a request by DrKonqi
-    } while (r == -1 && errno == EINTR);
-    // only continue if POLLIN event returned
-    if (r == 0) { // timeout
-        return 0;
-    } else if (r == -1 || !(fd.revents & POLLIN)) { // some error
-        return -1;
-    }
-
-    static struct sockaddr_un drkonqi_client;
-    static socklen_t cllength = sizeof(drkonqi_client);
-    int clsockfd;
-    do {
-        clsockfd = accept(sockfd, (struct sockaddr *)&drkonqi_client, &cllength);
-    } while (clsockfd == -1 && errno == EINTR);
-    if (clsockfd < 0) {
-        return -1;
-    }
-
-    // check whether the message is coming from DrKonqi
-    static struct ucred ucred;
-    static socklen_t credlen = sizeof(struct ucred);
-    if (getsockopt(clsockfd, SOL_SOCKET, SO_PEERCRED, &ucred, &credlen) < 0) {
-        return -1;
-    }
-
-    if (ucred.pid != pid) {
-        fprintf(stderr, "Warning: peer pid does not match DrKonqi pid\n");
-        return -1;
-    }
-
-    // read PID to change ptrace scope
-    static const int msize = 21; // most digits in a 64bit int (+sign +'\0')
-    char msg[msize];
-    if (read_socket(clsockfd, msg, msize) == 0) {
-        int dpid = atoi(msg);
-        prctl(PR_SET_PTRACER, dpid, 0, 0, 0);
-        // confirm change to DrKonqi
-        if (write_socket(clsockfd, msg, msize) == 0) {
-            fprintf(stderr, "KCrash: ptrace access transferred to %s\n", msg);
-        }
-    }
-    close(clsockfd);
-
-    return 1;
-}
-
-#endif
-
 #endif // Q_OS_UNIX
 
 void KCrash::setErrorMessage(const QString &message)
 {
     s_kcrashErrorMessage.reset(qstrdup(message.toUtf8().constData()));
 }
+
+void KCrash::setErrorTags(const QHash<QString, QString> &details)
+{
+    DetailsHash data;
+    for (const auto &[key, value] : details.asKeyValueRange()) {
+        data.insert(("--"_L1 + key).toUtf8(), value.toUtf8());
+    }
+    // A bit awkard. We want the s_details to be const so we can't accidentally cause
+    // detachments, so we move our data into a unique ptr that is const.
+    s_tags = std::make_unique<const DetailsHash>(std::move(data));
+}
+
+void KCrash::setErrorExtraData(const QHash<QString, QString> &details)
+{
+    DetailsHash data;
+    for (const auto &[key, value] : details.asKeyValueRange()) {
+        data.insert(("--"_L1 + key).toUtf8(), value.toUtf8());
+    }
+    // A bit awkard. We want the s_details to be const so we can't accidentally cause
+    // detachments, so we move our data into a unique ptr that is const.
+    s_extraData = std::make_unique<const DetailsHash>(std::move(data));
+}
+
+void KCrash::setGPUData(const QVariantHash &gpuData)
+{
+    DetailsHash data;
+    for (const auto &[key, value] : gpuData.asKeyValueRange()) {
+        data.insert(("--"_L1 + key).toUtf8(), value.toByteArray());
+    }
+    // A bit awkard. We want the s_details to be const so we can't accidentally cause
+    // detachments, so we move our data into a unique ptr that is const.
+    s_gpuContext = std::make_unique<const DetailsHash>(std::move(data));
+}
